
%  UNCONDITIONAL GOAL IMPLICATION CHECK

:- import subsumes/2 from subsumes.
:- import member/2, length/2, append/3 from basics.

% axiom/2 : checks for semantic equivalence of two predicates 
%           based on the syntax of their clauses
% assume Pred1, Pred2 are of the form p/n 
% containing the predicate name and arity

% Test for syntactic goal equivalence
% Given a program P (which is assumed to exist as dynamic code, where
% the clauses of the program are asserted as :
%     clauseI(ClauseId, (Head :- Body), Annotation)
% the routine will check whether two given preds are equivalent in P.
% The defn of syntactic equivalence used here is given in TACAS '00.

% Complexity of the algorithm implemented here :
%     O(p^2 * n^2 * 2^b)
% where : p = Number of predicates in P_0
%         n = Max. number of clauses of any predicate in P_i
%         b = Max. number of body literals in any clause in P_i

cleanup_and_axiom_chk(Pred1, Pred2) :-
	find_clauses(Pred1, CL1),
	find_clauses(Pred2, CL2),
	cleanup_clauses(CL1, _), 
	cleanup_clauses(CL2, _), !,
	axiom(Pred1, Pred2).

not_axiom_chk(Pred1, Pred2) :-
	abolish_all_tables, not_axiom(Pred1, Pred2).
	
axiom(Pred1, Pred2) :-
	abolish_all_tables,                  % avoiding the use of tnot/1
	( not_axiom(Pred1, Pred2) -> fail    % prevents unnecessary completion
		                  ;  true
        ).
	
:- table not_axiom/2.

% not_axiom(P, Q) is true
% 	if P => Q cannot be shown from the syntax of their clauses.

not_axiom(P/M, Q/N) :-
	( proved(P/M, Q/N) -> fail
		           ;  ( M \= N -> true   % predicates of 
                                                 % different arity
                                       ;  not_ordered(P/M, Q/N)
                              )
        ).

not_ordered(P/N, Q/N) :-
	find_clauses(P/N, Cp),
	find_clauses(Q/N, Cq),
        no_clause_maps(Cp, Cq).

% no_clause_maps(Cp, Cq) is true if clause_sets Cp,Cq are such that
% from their syntax we *cannot* infer P => Q
% Case 1 : Cp = [] : Trivially we have P => Q, so we fail out
% Case 2 : Cp \= [], Cq = [] : 
%           Then we cannot construct a map Cp -> Cq, so we give up.
% Case 3 : Otherwise : Construct all possible maps Cp -> Cq and check.
           
no_clause_maps(Cp, Cq) :-
	Cp = [H|T],
	( all_not_ordered(H, Cq) -> true
			         ;  no_clause_maps(T, Cq)
        ).


:- index all_not_ordered/2-2.

all_not_ordered(_, []).
all_not_ordered(C, [D|Rest]) :-
	not_ordered_clause(C, D),
	all_not_ordered(C, Rest).

not_ordered_clause(C, D) :-
	get_clause(C, (Hd1 :- Body1)),
	get_clause(D, (Hd2 :- Body2)),
	exec_builtins(Body1, Bd1),
	exec_builtins(Body2, Bd2),
	length(Bd1, M), length(Bd2, N),
	( M \= N -> true 
                 ;  findall(X, permute(Bd2, X), Bd2List),
                    body_not_axiom(Hd1, Bd1, Hd2, Bd2List)
        ).
/*	
	( M \= N ; 
	  M == N, ( not(clause_inst_chk((Hd1 :- Bd1), (Hd2 :- Bd2)))
	            ; exists_not_axiom(Bd1, Bd2)
                  )
        ).
*/




% try out all permutations of the body, and show that all
% of them are inequivalent.
% we can later change this to try only 1 permutation
% thereby having a stronger definition of syntactic equivalence

:- index body_not_axiom/4-4.

body_not_axiom(_, _, _, []).
body_not_axiom(Hd1, Bd1, Hd2, [Bd2|Rest2]) :-
	( clause_inst_chk((Hd1 :- Bd1), (Hd2 :- Bd2))
		-> exists_not_axiom(Bd1, Bd2)
                ;  true
        ), body_not_axiom(Hd1,Bd1,Hd2,Rest2).

exists_not_axiom([A1|Rest1], [A2|Rest2]) :-
	functor(A1, P, M), functor(A2, Q, N),
	( not_axiom(P/M, Q/N) ;  exists_not_axiom(Rest1, Rest2)
        ).


% clause_inst_chk(C, D) 
% ---  true if C is an instance of D modulo predicate renaming
% ---  creates bindings. do a copy_term before calling if needed.

clause_inst_chk((H :- B), (Hd :- Bd)) :-
	pred_replace((H :- B), dummy, (H1 :- B1)),
	pred_replace((Hd :- Bd), dummy, (Hd1 :- Bd1)),
        subsumes((Hd1 :- Bd1), (H1 :- B1)).

pred_replace((H :- B), P, (H1 :- B1)) :-
	!, pred_replace(H, P, H1),  
	pred_replace(B, P, B1).
pred_replace((G1, G2), P, (H1, H2)) :-
	!, pred_replace(G1, P, H1),
	pred_replace(G2, P, H2).
pred_replace([G1|G2], P, [H1|H2]) :-
	!, pred_replace(G1, P, H1),
	pred_replace(G2, P, H2).
pred_replace([], _, []) :- !.
pred_replace(G, P, H) :-
	G =.. [_|X], H =.. [P|X].


gen_index(Lo, Hi, X) :- 
	Hi >= Lo, 
	( X = Lo ; (Lo1 is Lo +1, gen_index(Lo1, Hi, X))
	).
	
%-----------------------------------------------------------------

exec_builtins(Bd, Bd1) :-
	( execute_builtins(Bd, X) -> Bd1 = X
		                  ;  Bd1 = [fail]
        ).