
:- import length/2, member/2, append/3 from basics.
:- import delete_ith/4 from listutil.
:- import 't not'/1, table_state/2 from tables.
:- import variant/2 from subsumes.

:- dynamic inductive/1.
:- dynamic unfold_prohibited/2.
:- dynamic unf_beyond_depth/1.
%-------------------------------------------------------------------

nonconsumer(ClauseId, Bd) :- 
	history(ClauseId, unfold, _), open_atoms(Bd).

open_atoms([]).
open_atoms([A|Rest]) :- open_atom(A), open_atoms(Rest).

%------------------------------------------------------

unfold(ClauseId, _) :-
	get_clause(ClauseId, (_ :- Bd), _),
	unfold_flag(nonconsumer_freeze), 
	nonconsumer(ClauseId, Bd), !, fail.

unfold(ClauseId, NewClauses) :-
	select(ClauseId, Num), 
	singleStepUnf(ClauseId, Num, NewClauses).

verbose_unfold(ClauseId, NewClauses) :-
	unfold(ClauseId, NewClauses), 
	write_clauses(NewClauses).

pos_unfold(ClauseId, Num, NewClauses) :-
	singleStepUnf(ClauseId,Num,NewClauses),
	write_clauses(NewClauses).

max_unfold(CList, UList) :-
	member(C,CList), select(C,Num), !,
	remove(C, CList, CList1),
	singleStepUnf(C, Num, CU),
	append(CList1, CU, UList1), 
	max_unfold(UList1, UList).

max_unfold(CList, CList).


%--------------------------------------------------------------
% Single step unfolding of a single body atom in the
% body of clause ClauseId.

singleStepUnf(ClauseId, Num, NewClauses) :-
	get_clause(ClauseId, (Hd :- Bd), _),
	get_body_atom((Hd :- Bd), Num, G),
	( predicate_property(G, built_in)
		-> unfold_builtin(ClauseId, Num, NewClauses)
	        ;  ( findall(UId, get_unifying_clause(G, UId), UList),
	             unfold_clauses(ClauseId, Num, UList, NewClauses)
                   )
	), retractall(clauseI(ClauseId, _, _)).
	%assert(old_clause(ClauseId, (Hd :- Bd))).

%------------------------------------------------------------

unfold_builtin(CId, Num, NewClauses) :-
	get_clause(CId, (Hd :- Bd), Annot),
	get_body_atom((Hd :- Bd), Num, G),
	( call(G) -> delete_ith(Num, Bd, G, BdRest),
		     gen_new_clauseNum(New), 
		     addAnnot(Annot, (1,1), A1),
		     assert_clause(New,(Hd :- BdRest),A1),
		     NewClauses = [New],
		     update_unfold_history(NewClauses,CId,Num,[builtin])
	          ;  NewClauses = [],
		     retractall(unfold_prohibited(CId, _))
        ).
                     


unfold_clauses(CId, Num, UList, NewClauses) :-
	unf_clauses(CId, Num, UList, NewClauses),
	update_unfold_history(NewClauses, CId, Num, UList).


unf_clauses(_ClauseId, _Num, [], []).

unf_clauses(ClauseId, Num, [UId|URest], [New|NewRest]) :-
	get_clause(ClauseId, C, A),
	get_clause(UId, C0, A0),
	gen_new_clauseNum(New),
        resolveAnswerClause(C, Num, C0, C1),
        addAnnot(A, A0, A1),
        assert_clause(New, C1, A1), 
        %nl, write_clause(New, C1, A1), 
	unf_clauses(ClauseId, Num, URest, NewRest).
	

update_unfold_history([], CId, _, []) :- 
	retractall(unfold_prohibited(CId, _)).

update_unfold_history([N|Rest], CId, Num, [UId|Urest]) :-
	update_unf_history(N, CId, Num, UId),
	update_unfold_history(Rest, CId, Num, Urest).

update_unf_history(N, ClauseId, Num, UId) :-
	get_clause(ClauseId, C, _),
	get_body_atom(C, Num, G),
	assert(history(N, unfold, [(ClauseId, G, UId)])),
	( unfold_flag(suspend_expand) 
		-> true
		;  unfold_flag(suspend_freeze),
		   update_prohib(N, ClauseId, Num, UId)
        ).


get_unifying_clause(G, UId) :-
	get_clause(UId, (H :- _), _), unifies(G, H).

%===================================================================
% Lower level routines
%=================================================================== 

% Updation of marks about:
%       the unfolding of "which" atoms is prohibited.

update_prohib(N, CId, Num, UId) :- 
	get_clause(N, (_ :- Body), _),
	new_prohib(Body, 1, N),
	propagate_prohib(CId, Num, UId, N).

new_prohib([], _Pos, _N).

new_prohib([G|Rest], Pos, N) :-
	( inductive(G), exists_improper_anc(G, N) 
		-> assert(unfold_prohibited(N, Pos))
		;  true
        ), Pos1 is Pos+1, new_prohib(Rest, Pos1, N).

propagate_prohib(CId, Num, UId, N) :-
	unfold_prohibited(CId, Pos),
	position_arith(Pos, Num, UId, New_pos),
	assert(unfold_prohibited(N, New_pos)),
	fail.

propagate_prohib(_, _, _, _).

position_arith(Pos, Num, builtin, Pos1) :-
	!, ( Pos < Num -> Pos1 = Pos 
		       ;  Pos1 is Pos - 1
        ).
position_arith(Pos, Num, _, Pos) :- Pos < Num, !.
position_arith(Pos, _Num, UId, NewPos) :-
	clauseI(UId, (_ :- Bd), _),
	length(Bd, Len), NewPos is Pos+Len-1.


%----------------------------------------------------------------------
%
% resolveAnswerClause(C1, N, C2, C) 
%     : C1 is a program clause
%     : C2 is an answer clause (H :- B) 
%     : resolve Nth body atom of C1 using C2 to produce clause C 
%     : which is a program clause.

resolveAnswerClause((H :- Bd), Num, (G :- B), Clause) :-
	get_body_atom((H :- Bd), Num, G), % implicit unification
	N is Num-1, split_list(Bd,N, Pre, [G|Post]),
        append(Pre,B,I), append(I, Post, Body),
        Clause = (H :- Body).


get_body_atom((_ :- B), N, G) :- find_pos(B, N, G).
/*
	( is_list(B) -> find_pos(B, N, G)
	             ;  commaConvert(B, B1),
		        find_pos(B1, N, G)
        ).
*/

% DEAD CODE
/*
resolveOnce(C, (G :- B), Clause) :-
        C = (H :- (Neck, Bd)),
        findGinBody(G,Bd,Pre,Post),      % Unification of G is implicit
        append(Pre, B, I), append(I, Post, Body),
        Clause = (H :- (Neck,Body)).
*/

%-----------------------------------------------------------------

% selection function for selecting a body atom to unfold.

select(ClauseId, Num) :-
	get_clause(ClauseId, (_ :- Body), _),
	select(ClauseId, Body, 1, Num).

select(CId, [G|Rest], ANum, Num) :-
	( select_strategy(G, ANum, CId)
		-> Num = ANum
		;  Anum1 is ANum + 1, 
		   select(CId, Rest, Anum1, Num)
	).	

%---------------------------------------------------------------------
% select_strategy/2 will  be programmed by user.
% some routines for letting the user program select_strategy/2.

depth_check(G) :-
	unf_beyond_depth(G1), my_subsumes(G1, G), !.
depth_check(G) :-	
	max_arg_depth(G, N),  % depth of any argument is \leq N
	maxDepthConstant(K), N < K.

consumer_chk(G) :- 
	functor(G, F, N), 
	( consumer(F/N) -> \+ open_atom(G)
			;  true
        ).

depth_and_consumer_chk(G) :-
	functor(G, F, N),
	( consumer(F/N) -> \+ open_atom(G)
			;  depth_check(G)
	).
	

always_chk(G) :- functor(G, F, N), always_unfold(F/N).
			  % i.e. always unfold this predicate occurrence

ancestor_check(G, _) :- 
	\+ inductive(G).
ancestor_check(G, CId) :- 
	inductive(G), \+ exists_improper_anc(G, CId).

exists_improper_anc(G, CId) :-
	anc(CId, G1), variant(G, G1).

anc(CId, G) :- anc(CId, _C, G).

anc(C1, CA, G) :- parent(C1, CA, G).
anc(C1, CA, G) :- parent(C1, CP, _), anc(CP, CA, G).

parent(C1, CP, G) :- history(C1, unfold, [(CP, G, _)]).


sharedVar_check(G, CId) :-
	get_clause(CId, (_ :- Body), _),
	\+ exists_improper_sibling(G, CId, Body).

exists_improper_sibling(G, CId, Body) :-
	another_atom(Body, G, G1), shared_var(G, G1, X), 
        maxdepth(X, G, N), maxdepth(X, G1, N1), N1 >= N,
	exists_improper_anc(G1, CId),
	%exists_subsumed_anc(G, CId),
	functor(G, F, Num), pred_tc(F/Num, F/Num).

exists_subsumed_anc(G, CId) :-
	anc(CId, Ga), subs_check(G, Ga).

%----------------------------------------------------------------

gen_new_node(N, Clause,Ctr) :- 
	gen_new_nodeNum(N), assert(ufTreeNode(N,Clause,Ctr)).

addAnnot((A1,A2), (B1,B2), (C1,C2)) :-
	plus_annot(A1, B1, C1), plus_annot(A2, B2, C2).


%====================================================================

